package de.fraunhofer.iem.dataprovider.tools.occmd.service

import de.fraunhofer.iem.dataprovider.configuration.DirectoryPathsProperties
import de.fraunhofer.iem.dataprovider.configuration.OpenCodeGitlabApiProperties
import de.fraunhofer.iem.dataprovider.logger.getLogger
import de.fraunhofer.iem.dataprovider.tool.dto.CreateToolDto
import de.fraunhofer.iem.dataprovider.tool.entity.ToolEntity
import de.fraunhofer.iem.dataprovider.tool.service.ToolService
import de.fraunhofer.iem.dataprovider.tools.occmd.json.RawResultJson
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.coroutineScope
import kotlinx.coroutines.future.await
import kotlinx.coroutines.withContext
import kotlinx.serialization.json.Json
import org.eclipse.jgit.api.Git
import org.springframework.stereotype.Service
import java.io.File
import java.nio.file.Files.isExecutable
import java.nio.file.Paths
import java.util.*
import kotlin.io.path.ExperimentalPathApi
import kotlin.io.path.deleteRecursively

@Service
class OccmdService(
    private val dirProperties: DirectoryPathsProperties,
    private val gitlabApiProperties: OpenCodeGitlabApiProperties,
    private val toolService: ToolService
) {

    val logger = getLogger(javaClass)

    val toolEntity: ToolEntity = getOrCreateToolEntity()

    private final fun getOrCreateToolEntity(): ToolEntity {
        val createToolDto = CreateToolDto("OCCMD")
        return toolService.findOrCreateTool(createToolDto)
    }


    /**
     * First, we clone the given repository from opencode to the path provided in the application.properties.
     *
     * Next, we run the OCCMD tool in a separate process using Java's process builder.
     *
     * Lastly, after calculating the KPIs of the OCCMD tool run, we delete the cloned repository and return
     * the DTOs.
     */
    @OptIn(ExperimentalPathApi::class)
    suspend fun runOccmd(repoId: Long, repoUrl: String): List<RawResultJson> {

        val rawOccmdResults = try {
            // clone repo
            val outDir = Paths.get(dirProperties.gitCloneTargetDirectory, "$repoId-${Date().time}")

            cloneGit(repoUrl, outDir.toFile())
            // run tool
            val occmdResult = runOccmd(
                dirProperties.occmdPath, arrayOf(
                    outDir.toString(), "$repoId",
                    gitlabApiProperties.accessToken
                )
            )
            // delete cloned repo
            try {
                outDir.deleteRecursively()
            } catch (e: Exception) {
                logger.error("Delete directory after occmd run failed for dir $outDir with $e")
            }

            occmdResult
        } catch (e: Exception) {
            logger.error("OCCMD tool run failed with $e")
            emptyList<RawResultJson>()
        }
        // TODO: right now we fire and forget, in a sense that we run the tool calculate the
        // KPIs and forget the tool results. For this tool we want to manually store the results
        // until it is included into the official CI/CD pipeline.

        return rawOccmdResults
    }


    private suspend fun runOccmd(execPath: String, flags: Array<String>): List<RawResultJson> = coroutineScope {
        val toolResults = mutableListOf<RawResultJson>()
        if (isExecutable(execPath)) {

            val process = withContext(Dispatchers.IO) {
                ProcessBuilder(execPath, *flags).start()
            }
            process.onExit().await()
            val json = Json { ignoreUnknownKeys = true }
            process.inputStream.bufferedReader().forEachLine {
                toolResults.add(json.decodeFromString(it))
            }
        } else {
            logger.error("Given execPath is not an executable $execPath.")
        }

        return@coroutineScope toolResults
    }

    private fun isExecutable(filePath: String): Boolean {
        val path = Paths.get(filePath)
        return isExecutable(path)
    }

    private suspend fun cloneGit(repoUrl: String, outDir: File) {
        val git: Git = Git.cloneRepository()
            .setURI(repoUrl)
            .setDirectory(outDir)
            .call()
        git.close()
    }
}
